@@ -1,5 +1,5 @@
 """
- * $Id: 1.1+11 utils.py de8100d559ad 2021-06-25 od $
+ * $Id: 1.1+16 utils.py f53fd7143efa 2022-02-17 od $
  *
  * This file is part of the Cloud Services Integration Platform (CSIP),
  * a Model-as-a-Service framework, API and application suite.
@@ -10,6 +10,7 @@
  * See the LICENSE file in the project root for more information.
 """
 
+import fnmatch
 from os import path
 from pathlib import Path
 from shutil import make_archive
@@ -19,6 +20,7 @@ from requests.exceptions import Timeout
 
 import json, os, re, requests, sys, time, glob
 
+
 class Client(object):
     """
     CSIP client class.
@@ -67,8 +69,8 @@ class Client(object):
 
     # CONFIG defaults
     CONN_TIMEOUT = 2  # for "conn_timeout"
-    READ_TIMEOUT = 10 # for "read_timeout"
-    SERVICE_TIMEOUT = 30 # for "service_timeout"
+    READ_TIMEOUT = 10  # for "read_timeout"
+    SERVICE_TIMEOUT = 30  # for "service_timeout"
     RETRY = 3  # for "retry"
     ALLOW_REDIRECTS = True  # for "allow_redirects"
 
@@ -105,7 +107,6 @@ class Client(object):
             data = json.load(f)
         return data
 
-
     @staticmethod
     def save_json(file: str, js: Dict) -> None:
         """Save a dict to a file"""
@@ -117,6 +118,11 @@ class Client(object):
         """Create dict key:name, value: whole json"""
         return {i[Client.KEY_NAME]: i for i in json}
 
+    @staticmethod
+    def __create_list(json: Dict) -> List[Dict]:
+        """Create dict key:name, value: whole json"""
+        return [v for v in json.values()]
+
     @staticmethod
     def __remove(d: Dict, names: List[str]) -> None:
         """ Removes entries from a dict"""
@@ -140,6 +146,9 @@ class Client(object):
             mi = {}
             data = None
 
+        if response is not None:
+            response.close()
+
         # if not section in response_json:
         #     data = {}
         # else:
@@ -153,7 +162,8 @@ class Client(object):
     def __get0(url: str, conf: Dict):
         """HTTP GET to generate a Client"""
         conf = conf or {}
-        timeout = (conf.get('http_conn_timeout', Client.CONN_TIMEOUT), conf.get('http_read_timeout', Client.READ_TIMEOUT))
+        timeout = (
+        conf.get('http_conn_timeout', Client.CONN_TIMEOUT), conf.get('http_read_timeout', Client.READ_TIMEOUT))
         retry = conf.get('http_retry', Client.RETRY)
         allow_redirects = conf.get('http_allow_redirects', Client.ALLOW_REDIRECTS)
 
@@ -174,12 +184,18 @@ class Client(object):
                 ## we will try again
                 retry += -1
                 http_status = -1
-                # print("Timeout Error:", errt)
+                print("Timeout Error:", errt)
             except requests.exceptions.ConnectionError as errc:
                 retry += -1
                 http_status = -2
+                template = "Exception {0} occurred. Arguments:\n{1!r}"
+                message = template.format(type(errc).__name__, errc.args)
+                print("ConnectionError: ", message)
                 # print("Error Connecting:", errc)
             except requests.exceptions.RequestException as err:
+                template = "Exception {0} occurred. Arguments:\n{1!r}"
+                message = template.format(type(err).__name__, err.args)
+                print("RequestException: ", message)
                 http_status = -3
                 break
         return response, http_status
@@ -189,9 +205,11 @@ class Client(object):
         """HTTP POST to generate a Client"""
         conf = conf or {}
         if sync:
-            timeout = (conf.get('http_conn_timeout', Client.CONN_TIMEOUT), conf.get('service_timeout', Client.SERVICE_TIMEOUT))
+            timeout = (
+            conf.get('http_conn_timeout', Client.CONN_TIMEOUT), conf.get('service_timeout', Client.SERVICE_TIMEOUT))
         else:
-            timeout = (conf.get('http_conn_timeout', Client.CONN_TIMEOUT), conf.get('http_read_timeout', Client.READ_TIMEOUT))
+            timeout = (
+            conf.get('http_conn_timeout', Client.CONN_TIMEOUT), conf.get('http_read_timeout', Client.READ_TIMEOUT))
 
         retry = conf.get('http_retry', Client.RETRY)
         allow_redirects = conf.get('http_allow_redirects', Client.ALLOW_REDIRECTS)
@@ -223,6 +241,8 @@ class Client(object):
             except requests.exceptions.RequestException as err:
                 http_status = -3
                 break
+            if response is not None:
+                response.close()
         return response_json, http_status
 
     @staticmethod
@@ -253,13 +273,31 @@ class Client(object):
 
         response, http_status = Client.__get0(url, conf)
         if http_status != 200:
-            raise Exception("HTTP Error " + http_status)
+            raise Exception("HTTP Error: {}".format(http_status))
 
         response_json = response.json()
+        if response is not None:
+            response.close()
+
         return [Client(metainfo=service) for service in response_json]
 
     @staticmethod
-    def get_capabilities(c, conf: Dict = None) -> "Client":
+    def list_capabilities(urls: List['str'], conf: Dict = None) -> Dict:
+        """ Get a combined list of services from various endpoints as json"""
+        res = []
+        for url in urls:
+            response, http_status = Client.__get0(url, conf)
+            if http_status != 200:
+                raise Exception("HTTP Error: {}".format(http_status))
+
+            response_json = response.json()
+            res.extend(response_json)
+            if response is not None:
+                response.close()
+        return res
+
+    @staticmethod
+    def get_capabilities(c, conf: Dict = None) -> 'Client':
         """Get the parameter of a service. Creates a request Client"""
         if isinstance(c, str):
             return Client.__get(c, conf=conf)
@@ -274,13 +312,13 @@ class Client(object):
         raise Exception("Illegal argument type: " + type(c))
 
     def save_to(self, file: str) -> None:
-        """Publish/subscribe execution """
-        d = { Client.METAINFO: self.metainfo }
+        """Saves the Client payload (either request or response)"""
+        d = {Client.METAINFO: self.metainfo}
         if self.parent:
-            d[Client.PARAMETER] = self.parent.data
-            d[Client.RESULT] = self.data
+            d[Client.PARAMETER] = Client.__create_list(self.parent.data)
+            d[Client.RESULT] = Client.__create_list(self.data)
         else:
-            d[Client.PARAMETER] = self.data
+            d[Client.PARAMETER] = Client.__create_list(self.data)
 
         """Save the file"""
         with open(file, 'w') as outfile:
@@ -469,7 +507,6 @@ class Client(object):
 
     def download_data_files(self, files: List[str] = None, dir: str = ".", prefix: str = '', conf: Dict = None) -> None:
         """ Downloads file(s) if this Client object is a valid response and a query url as value.
-            If no 'file' argument is provided, all files will be downloaded.
             If no 'dir' is provided, the current directory will be used.
         """
         if files is None or not files:
@@ -483,19 +520,28 @@ class Client(object):
         if suid is None:
             raise Exception("No suid.")
 
+        dnames = self.get_data_names()
+        suid = self.get_suid()
+
         print("downloading to '" + dir + "':")
         for name in files:
-            if name not in self.data:
-                raise Exception("Not in data: " + name)
-            val = self.get_data_value(name)
-            print("  --> " + name)
-            response, http_status = Client.__get0(val, conf)
-            if http_status == 200:
-                with open(os.path.join(dir, prefix + name), 'wb') as outfile:
-                    outfile.write(response.content)
-            else:
-                raise Exception("Error downloading : " + val)
-
+            # if name not in self.data:
+            #     raise Exception("Not found for download: " + name)
+            for dname in dnames:
+                if fnmatch.fnmatch(dname, name):
+                    val = self.get_data_value(dname)
+                    if isinstance(val, str) and val.endswith('/q/' + suid + '/' + dname):
+                        print("  --> " + dname)
+                        response, http_status = Client.__get0(val, conf)
+                        if http_status == 200:
+                            with open(os.path.join(dir, prefix + dname), 'wb') as outfile:
+                                outfile.write(response.content)
+                            if response is not None:
+                                response.close()
+                        else:
+                            if response is not None:
+                                response.close()
+                            raise Exception("Error (" + str(http_status) + ") downloading : " + val)
 
     #### pubsub related properties
 
@@ -580,6 +626,27 @@ class Client(object):
         """Print the metainfo  """
         return json.dumps(self.metainfo, indent=indent)
 
+    #### cosu
+    def add_cosu(self, name: str, of: str, data: List) -> None:
+        """Adds a cosu entry"""
+        entry = {'name': name, 'of': of, 'data': data}
+        self.metainfo.setdefault('cosu', []).append(entry)
+
+    def has_cosu(self) -> bool:
+        """Check if there is  cosu information in metainfo"""
+        return 'cosu' in self.metainfo
+
+    def get_cosu_names(self) -> List:
+        """Get all cosu names as list"""
+        cosu = self.metainfo.get('cosu', [])
+        return [i[Client.KEY_NAME]for i in cosu]
+
+    def get_cosu_value(self, name: str) -> float:
+        cosu = self.metainfo.get('cosu', [])
+        m = {i[Client.KEY_NAME]: i for i in cosu}
+        entry = m.get(name, {})
+        return entry.get(Client.KEY_VALUE, None)
+
     #### data
 
     def get_data_attr(self, name: str, key: str) -> str:
@@ -630,6 +697,10 @@ class Client(object):
         """Check if data entry exists """
         return name in self.data
 
+    def get_data_names(self) -> List[str]:
+        """Get all names"""
+        return self.data.keys()
+
     def get_data(self, name: str) -> Dict:
         """Get the whole entry as dict"""
         return self.data.get(name, {})
@@ -721,7 +792,6 @@ class Client(object):
 
     #### printing support
 
-
     def asJSON(self) -> Dict:
         # self.metainfo[self.META_KEY_MODE] = self.META_VAL_SYNC if sync else self.META_VAL_ASYNC
         request_json = {self.METAINFO: self.metainfo, self.PARAMETER: list(self.data.values())}
